''***********************************************
''*	PS/2 Keyboard Driver v1.0.1             *
''*	Author: Chip Gracey                     *
''*	Copyright (c) 2004 Parallax, Inc.	*
''*	See end of file for terms of use.	*
''***********************************************

{-----------------REVISION HISTORY-----------------
 v1.0.1 - Updated 6/15/2006 to work with Propeller Tool 0.96}

CON
	stat_extended	= 1		' extended keycode flag
	stat_break	= 2		' break code flag
	stat_config	= 4		' configure keyboard flag
	stat_reset	= 8		' reset keyboard flag

VAR
	long	cog
	long	par_states[8]	'key states (256 bits)	read-only
	long	par_present	'keyboard present	read-only

PUB start(dpin, cpin) : okay

'' Start keyboard driver - starts a cog
'' returns false if no cog available
''
''	dpin	= data signal on PS/2 jack
''	cpin	= clock signal on PS/2 jack
''
''	use 100-ohm resistors between pins and jack
''	use 10K-ohm resistors to pull jack-side signals to VDD
''	connect jack-power to 5V, jack-gnd to VSS
''
'' all lock-keys will be enabled, NumLock will be initially 'on',
'' and auto-repeat will be set to 15cps with a delay of .5s

	okay := startx(dpin, cpin, %0_000_100, %01_01000)

PUB startx(dpin, cpin, locks, auto) : okay

'' Like start, but allows you to specify lock settings and auto-repeat
''
''	locks = lock setup
''		bit 6 disallows shift-alphas (case set soley by CapsLock)
''		bits 5..3 disallow toggle of NumLock/CapsLock/ScrollLock state
''		bits 2..0 specify initial state of NumLock/CapsLock/ScrollLock
''		(eg. %0_001_100 = disallow ScrollLock, NumLock initially 'on')
''
''	auto	= auto-repeat setup
''		bits 6..5 specify delay (0=.25s, 1=.5s, 2=.75s, 3=1s)
''		bits 4..0 specify repeat rate (0=30cps..31=2cps)
''		(eg %01_00000 = .5s delay, 30cps repeat)

	stop
	longmove(@par_states, @dpin, 4)
	okay := cog := cognew(@entry, @par_states) + 1


PUB stop
'' Stop keyboard driver - frees a cog
	if cog
		cogstop(cog~ -	1)
	longfill(@par_states, 0, 4+1)


PUB present : truefalse
'' Check if keyboard present - valid ~2s after start
'' returns t|f
	truefalse := -par_present

PUB keystate(k) : state
'' Get the state of a particular key
'' returns t|f
	if k < 128
		state := -(par_states[k >> 5] >> k & 1)
	else
		state := 0

PUB states_ptr
	return @par_states	' returns the location of par_states longs

DAT

'******************************************
'* Assembly language PS/2 keyboard driver *
'******************************************

		org	0
'
'
' Entry
'
entry			movd	:par, #_dpin			' load input parameters _dpin/_cpin/_locks/_auto
			mov	x, par
			mov	y, #4
			mov	t, #0
:par			rdlong	0, x
			add	:par, dlsb
			wrlong	t, x				' clear the states
			add	x, #4
			djnz	y, #:par

			mov	dmask, #1			' set pin masks
			shl	dmask, _dpin			' 0..31 (ignores bit #5)
			mov	cmask, #1
			shl	cmask, _cpin			' 0..31 (ignores bit #5)

			test	_dpin, #$20		wc	' modify port registers within code
			muxc	_d1, dlsb
			muxc	_d2, dlsb
			muxc	_d3, #1
			muxc	_d4, #1
			test	_cpin, #$20		wc
			muxc	_c1, dlsb
			muxc	_c2, dlsb
			muxc	_c3, #1
'
'
' Reset keyboard
'
reset			mov	dira, #0			' reset directions
			mov	dirb, #0

			movd	:par, #_states			' reset output parameters _states[8]/_present
			mov	x, #8 + 1
:par			mov	0, #0
			add	:par, dlsb
			djnz	x, #:par

			mov	stat, #stat_reset		' set reset flag
'
'
' Update parameters
'
update			movd	:par, #_states			' update output _states[8]/_present
			mov	x, par
			mov	y, #8 + 1
:par			wrlong	0, x
			add	:par, dlsb
			add	x, #4
			djnz	y, #:par

			mov	x, _states			' get keys $20..$3f
			and	x, ctrl_alt_del
			cmp	x, ctrl_alt_del		wz
		if_nz	jmp	#:not_reset
			mov	x, #$80
			clkset	x				' reset Propeller

:not_reset		test	stat, #stat_reset wc		' if reset flag, transmit reset command
		if_c	mov	data, #$ff
		if_c	call	#transmit
'
' Get scancode
'
newcode
			mov	stat, #0			' reset state
:same
			call	#receive			' receive byte from keyboard
			cmp	data,#$83+1		wc	' scancode?

	if_nc		cmp	data, #$AA		wz	' powerup/reset?
	if_nc_and_z	jmp	#configure

	if_nc		cmp	data, #$E0		wz	' extended?
	if_nc_and_z	or	stat, #stat_extended
	if_nc_and_z	jmp	#:same

	if_nc		cmp	data, #$F0		wz	' released?
	if_nc_and_z	or	stat, #stat_break
	if_nc_and_z	jmp	#:same

	if_nc		jmp	#newcode			' unknown, ignore
'
'
' Translate scancode and enter into buffer
'
			test	stat, #stat_extended	wc	' lookup code with extended flag
			rcl	data, #1
			call	#look

			cmp	data, #$ff		wz	' if unknown, ignore
		if_z	jmp	#newcode

			mov	t, _states+3			' remember lock keys in _states

			mov	x, data				' set/clear key bit in _states
			shr	x, #5
			add	x, #_states
			movd	:reg, x
			mov	y, #1
			shl	y, data
			test	stat, #stat_break	wc
:reg			muxnc	0-0, y

			test	stat, #stat_config	wc	' if not configure flag, done
		if_nc	jmp	#update				' else configure to update leds
'
' Configure keyboard
'
configure
			mov	data, #$F3			' set keyboard auto-repeat
			call	#transmit
			mov	data, _auto
			and	data, #%11_11111
			call	#transmit

			mov	data, #$ED			' set keyboard lock-leds
			call	#transmit
			mov	data, _locks
			rev	data, #-3 & $1F
			test	data, #%100	wc
			rcl	data, #1
			and	data, #%111
			call	#transmit

			mov	x, _locks			' insert locks into _states
			and	x, #%111
			shl	_states+3, #3
			or	_states+3, x
			ror	_states+3, #3

			mov	_present, #1			' set _present

			jmp	#update				' done
'
' Lookup byte in table
'
look
			ror	data, #2			' perform lookup
			movs	:reg, data
			add	:reg, #table
			shr	data, #27
			mov	x, data
:reg			mov	data, 0-0
			shr	data, x
			jmp	#rand				' isolate byte
'
' Transmit byte to keyboard
'
transmit
_c1			or	dira, cmask			' pull clock low
			movs	napshr, #13			' hold clock for ~128us (must be >100us)
			call	#nap
_d1			or	dira, dmask			' pull data low
			movs	napshr, #18			' hold data for ~4us
			call	#nap
_c2			xor	dira, cmask			' release clock

			test	data, #$0FF		wc	' append parity and stop bits to byte
			muxnc	data, #$100
			or	data, dlsb

			mov	x, #10				' transmit 10 bits
transmit_bit
			call	#wait_c0			' wait until clock low
			shr	data, #1		wc	' output data bit
_d2			muxnc	dira, dmask
			mov	wcond, c1			' wait until clock high
			call	#wait
			djnz	x, #transmit_bit		' another bit?

			mov	wcond, c0d0			' wait until clock and data low
			call	#wait
			mov	wcond, c1d1			' wait until clock and data high
			call	#wait

			call	#receive_ack			' receive ack byte with timed wait
			cmp	data, #$FA		wz	' if ack error, reset keyboard
		if_nz	jmp	#reset

transmit_ret
			ret
'
' Receive byte from keyboard
'
receive			test	_cpin,#$20		wc	' wait indefinitely for initial clock low
			waitpne cmask, cmask
receive_ack
			mov	x, #11				' read 11 bits
receive_bit
			call	#wait_c0			' wait until clock low
			movs	napshr, #16			' pause ~16us
			call	#nap
_d3			test	dmask, ina		wc	' input data bit
			rcr	data,#1
			mov	wcond, c1			' wait until clock high
			call	#wait
			djnz	x,#receive_bit			' another bit?

			shr	data, #22			' align byte
			test	data, #$1FF		wc	' if parity error, reset keyboard
		if_nc	jmp	#reset
rand			and	data, #$FF			' isolate byte

look_ret
receive_ack_ret
receive_ret
			ret
'
'
' Wait for clock/data to be in required state(s)
'
wait_c0			mov	wcond, c0			' (wait until clock low)
wait			mov	y, tenms			' set timeout to 10ms
wloop			movs	napshr, #18			' nap ~4us
			call	#nap
_c3			test	cmask, ina		wc	' check required state(s)
_d4			test	dmask, ina		wz	' loop until got state(s) or timeout
wcond	if_never	djnz	y, #wloop			' (replaced with c0/c1/c0d0/c1d1)
			tjz	y, #reset			' if timeout, reset keyboard
wait_ret
wait_c0_ret
			ret


c0	if_c		djnz	y, #wloop			' (if_never replacements)
c1	if_nc		djnz	y, #wloop
c0d0	if_c_or_nz	djnz	y, #wloop
c1d1	if_nc_or_z	djnz	y, #wloop
'
'
' Nap
'
nap			rdlong	t, #0				' get clkfreq
napshr			shr	t, #18/16/13			' shr scales time
			min	t, #3				' ensure waitcnt won't snag
			add	t, cnt				' add cnt to time
			waitcnt t, #0				' wait until time elapses (nap)

nap_ret
			ret
'
'
' Initialized data
'
'
dlsb			long	1 << 9
tenms			long	10_000 / 4
ctrl_alt_del		long	$80 << 24 | $80 << 8 | $80
'
' bit input
'  7  unused
'  6  Keyboard S6
'  5  Keyboard S3
'  4  Keyboard S5
'  3  Keyboard S4
'  2  Keyboard S0
'  1  Keyboard S2
'  0  Keyboard S1
'
'
'	; keyboard table
'	s rrrr ccc
'	-------------------------------
'	0_0001_000	#08 backspace
'	0_0001_001	#0d cr
'	0_0001_010	#2d -
'	0_0001_011	-----
'	0_0001_100	-----
'	0_0001_101	#40 @
'	0_0001_110	#17 ch (ctrl)
'
'	0_0010_000	#43 g
'	0_0010_001	#42 v
'	0_0010_010	#34 4
'	0_0010_011	#43 c
'	0_0010_100	#20 space
'	0_0010_101	#42 r
'	0_0010_110	#41 ]
'
'	0_0011_000	#3a :
'	0_0011_001	#2f /
'	0_0011_010	#30 0
'	0_0011_011	#32 2
'	0_0011_100	#31 1
'	0_0011_101	#50 p
'	0_0011_110	#43 [
'
'	0_0100_000	#3b ;
'	0_0100_001	#2e .
'	0_0100_010	#39 9
'	0_0100_011	#33 3
'	0_0100_100	#41 q
'	0_0100_101	#43 o
'	0_0100_110	-----
'
'	0_0101_000	#4c l
'	0_0101_001	#2c ,
'	0_0101_010	#38 8
'	0_0101_011	#43 w
'	0_0101_100	#41 a
'	0_0101_101	#41 i
'	0_0101_110	#12 [right]
'
'	0_0110_000	#43 k
'	0_0110_001	#41 m
'	0_0110_010	#37 7
'	0_0110_011	#41 e
'	0_0110_100	#43 s
'	0_0110_101	#41 u
'	0_0110_110	#14 [dn]
'
'	0_0111_000	#42 j
'	0_0111_001	#42 n
'	0_0111_010	#36 6
'	0_0111_011	#44 d
'	0_0111_100	#42 z
'	0_0111_101	#41 y
'	0_0111_110	#11 [left]
'
'	0_1000_000	#48 h
'	0_1000_001	#42 b
'	0_1000_010	#35 5
'	0_1000_011	#42 f
'	0_1000_100	#58 x
'	0_1000_101	#54 t
'	0_1000_110	#13 [up]
'
'	1_0001_000	#0c cs (clear screen)
'	1_0001_001	#1b esc
'	1_0001_010	#3d =
'	1_0001_011	-----
'	1_0001_100	-----
'	1_0001_101	-----
'	1_0001_110	#0a line feed
'
'	1_0010_000	-----
'	1_0010_001	-----
'	1_0010_010	#24 $
'	1_0010_011	-----
'	1_0010_100	-----
'	1_0010_101	-----
'	1_0010_110	#43 _
'
'	1_0011_000	#2a *
'	1_0011_001	#3f ?
'	1_0011_010	#42 ^
'	1_0011_011	#22 "
'	1_0011_100	#21 !
'	1_0011_101	-----
'	1_0011_110	#5c \
'
'	1_0100_000	#2b +
'	1_0100_001	#3e >
'	1_0100_010	#29 )
'	1_0100_011	#23 #
'	1_0100_100	-----
'	1_0100_101	-----
'	1_0100_110	-----
'
'	1_0101_000	-----
'	1_0101_001	#3c <
'	1_0101_010	#28 (
'	1_0101_011	-----
'	1_0101_100	-----
'	1_0101_101	-----
'	1_0101_110	#16 [right]
'
'	1_0110_000	-----
'	1_0110_001	-----
'	1_0110_010	#27 '
'	1_0110_011	-----
'	1_0110_100	-----
'	1_0110_101	-----
'	1_0110_110	-----
'
'	1_0111_000	-----
'	1_0111_001	-----
'	1_0111_010	#26 &
'	1_0111_011	-----
'	1_0111_100	-----
'	1_0111_101	-----
'	1_0111_110	#15 [left]
'
'	1_1000_000	-----
'	1_1000_001	-----
'	1_1000_010	#25 %
'	1_1000_011	-----
'	1_1000_100	-----
'	1_1000_101	-----
'	1_1000_110	-----
'
'	1_1111_111	#00
'	1_1111_111	#01
'	1_1111_111	#02
'	1_1111_111	#03
'	1_1111_111	#04
'	1_1111_111	#05
'	1_1111_111	#06
'	1_1111_111	#07
'	1_1111_111	#09
'	1_1111_111	#0b
'	1_1111_111	#0e
'	1_1111_111	#0f
'	1_1111_111	#10
'	1_1111_111	#18
'	1_1111_111	#19
'	1_1111_111	#1a
'	1_1111_111	#1c
'	1_1111_111	#1d
'	1_1111_111	#1e
'	1_1111_111	#1f
'
'***************************************************************************
'*   w/o SHIFT                             with SHIFT
'*   +-------------------------------+     +-------------------------------+
'*   | 0   1   2   3   4   5   6   7 |     | 0   1   2   3   4   5   6   7 |
'*+--+---+---+---+---+---+---+---+---+  +--+---+---+---+---+---+---+---+---+
'*|D0|BSP| H | J | K | L | ; | : | G |  |D0|BSP| h | j | k | l | + | * | g |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|D1|ENT| B | N | M | , | . | / | V |  |D1|ENT| b | n | m | < | > | ? | v |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|D2| - | 5 | 6 | 7 | 8 | 9 | 0 | 4 |  |D2| = | % | & | ' | ( | ) | ^ | $ |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|D3|   | F | D | E | W | 3 | 2 | C |  |D3|   | f | d | e | w | # | " | c |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|D4|SHF| X | Z | S | A | Q | 1 |SPC|  |D4|SHF| x | z | s | a | q | ! |SPC|
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|D5| @ | T | Y | U | I | O | P | R |  |D5| ` | t | y | u | i | o | p | r |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|D6|CH |UP |LFT|DN |RGT|   | [ | ] |  |D6|LF |UP |LFT|DN |RGT|   | \ | _ |
'*+--+---+---+---+---+---+---+---+---+  +--+---+---+---+---+---+---+---+---+
'*|D7|CTL|ALT|INS|DEL|PUP|PDN|HOM|END|  |D7|CTL|ALT|INS|DEL|PUP|PDN|HOM|END|
'*+--+---+---+---+---+---+---+---+---+  +--+---+---+---+---+---+---+---+---+
'***************************************************************************
'
'
' Lookup table
'									scan	extkey		regkey	()=keypad
'
table			word	($ff << 8) | $ff	'00
			word	($ff << 8) | $ff	'01			F9
			word	($ff << 8) | $ff	'02
			word	($ff << 8) | $ff	'03			F5
			word	($ff << 8) | $ff	'04			F3
			word	($ff << 8) | $ff	'05			F1
			word	($ff << 8) | $ff	'06			F2
			word	($ff << 8) | $ff	'07			F12
			word	($ff << 8) | $ff	'08
			word	($ff << 8) | $ff	'09			F10
			word	($ff << 8) | $ff	'0A			F8
			word	($ff << 8) | $ff	'0B			F6
			word	($ff << 8) | $ff	'0C			F4
			word	($ff << 8) | $ff	'0D			Tab
			word	($ff << 8) | (0*8+5)	'0E			`
			word	($ff << 8) | $ff	'0F
			word	($ff << 8) | $ff	'10	[WWW search]
			word	((1*8+7)<<8) | (1*8+7)	'11	Alt-R		Alt-L
			word	($ff << 8) | (0*8+4)	'12			Shift-L
			word	($ff << 8) | $ff	'13
			word	((0*8+7)<<8) | (0*8+7)	'14	Ctrl-R		Ctrl-L
			word	($ff << 8) | (5*8+4)	'15	[Prev Track]	q
			word	($ff << 8) | (6*8+4)	'16			1
			word	($ff << 8) | $ff	'17
			word	($ff << 8) | $ff	'18	[WWW favorites]
			word	($ff << 8) | $ff	'19
			word	($ff << 8) | (2*8+4)	'1A			z
			word	($ff << 8) | (3*8+4)	'1B			s
			word	($ff << 8) | (4*8+4)	'1C			a
			word	($ff << 8) | (4*8+3)	'1D			w
			word	($ff << 8) | (6*8+3)	'1E			2
			word	($ff << 8) | $ff	'1F	Win-L
			word	($ff << 8) | $ff	'20	[WWW refresh]
			word	($ff << 8) | (7*8+3)	'21	[Volume -]	c
			word	($ff << 8) | (1*8+4)	'22			x
			word	($ff << 8) | (2*8+3)	'23	[Mute]		d
			word	($ff << 8) | (3*8+3)	'24			e
			word	($ff << 8) | (7*8+2)	'25			4
			word	($ff << 8) | (5*8+3)	'26			3
			word	($ff << 8) | $ff	'27	Win-R
			word	($ff << 8) | $ff	'28	[WWW stop]
			word	($ff << 8) | (7*8+4)	'29			Space
			word	($ff << 8) | (7*8+1)	'2A			v
			word	($ff << 8) | (1*8+3)	'2B	[Calculator]	f
			word	($ff << 8) | (1*8+5)	'2C			t
			word	($ff << 8) | (7*8+5)	'2D			r
			word	($ff << 8) | (1*8+2)	'2E			5
			word	($ff << 8) | $ff	'2F	Apps
			word	($ff << 8) | $ff	'30	[WWW forward]
			word	($ff << 8) | (2*8+1)	'31			n
			word	($ff << 8) | (1*8+1)	'32	[Volume +]	b
			word	($ff << 8) | (1*8+0)	'33			h
			word	($ff << 8) | (7*8+0)	'34	[Play/Pause]	g
			word	($ff << 8) | (2*8+5)	'35			y
			word	($ff << 8) | (2*8+2)	'36			6
			word	($ff << 8) | $ff	'37	Power
			word	($ff << 8) | $ff	'38	[WWW back]
			word	($ff << 8) | $ff	'39
			word	($ff << 8) | (3*8+1)	'3A	[WWW home]	m
			word	($ff << 8) | (2*8+0)	'3B	[Stop]		j
			word	($ff << 8) | (3*8+5)	'3C			u
			word	($ff << 8) | (3*8+2)	'3D			7
			word	($ff << 8) | (4*8+2)	'3E			8
			word	($ff << 8) | $ff	'3F	Sleep
			word	($ff << 8) | $ff	'40	[My Computer]
			word	($ff << 8) | (4*8+1)	'41			,
			word	($ff << 8) | (3*8+0)	'42			k
			word	($ff << 8) | (4*8+5)	'43			i
			word	($ff << 8) | (5*8+5)	'44			o
			word	($ff << 8) | (6*8+2)	'45			0
			word	($ff << 8) | (5*8+2)	'46			9
			word	($ff << 8) | $ff	'47
			word	($ff << 8) | $ff	'48	[Email]
			word	($ff << 8) | (5*8+1)	'49			.
			word	((6*8+1)<<8) | (6*8+1)	'4A	(/)		/
			word	($ff << 8) | (4*8+0)	'4B			l
			word	($ff << 8) | (5*8+0)	'4C			;
			word	($ff << 8) | (6*8+5)	'4D	[Next Track]	p
			word	($ff << 8) | (0*8+2)	'4E			-
			word	($ff << 8) | $ff	'4F
			word	($ff << 8) | $ff	'50	[Media Select]
			word	($ff << 8) | $ff	'51
			word	($ff << 8) | (6*8+0)	'52			'
			word	($ff << 8) | $ff	'53
			word	($ff << 8) | (6*8+6)	'54			[
			word	($ff << 8) | (0*8+2)	'55			=		?????
			word	($ff << 8) | $ff	'56
			word	($ff << 8) | $ff	'57
			word	($ff << 8) | $ff	'58			CapsLock
			word	($ff << 8) | (0*8+4)	'59			Shift-R
			word	((0*8+1)<<8) | (0*8+1)	'5A	(Enter)		Enter
			word	($ff << 8) | (7*8+6)	'5B			]
			word	($ff << 8) | $ff	'5C
			word	($ff << 8) | (5*8+6)	'5D			\		?????
			word	($ff << 8) | $ff	'5E	WakeUp
			word	($ff << 8) | $ff	'5F
			word	($ff << 8) | $ff	'60
			word	($ff << 8) | $ff	'61
			word	($ff << 8) | $ff	'62
			word	($ff << 8) | $ff	'63
			word	($ff << 8) | $ff	'64
			word	($ff << 8) | $ff	'65
			word	($ff << 8) | (0*8+0)	'66			BackSpace
			word	($ff << 8) | $ff	'67
			word	($ff << 8) | $ff	'68
			word	((7*8+7)<<8) | $ff	'69	End			(1)
			word	($ff << 8) | $ff	'6A
			word	((2*8+6)<<8) | (2*8+6)	'6B	Left			(4)
			word	((6*8+7)<<8) | $ff	'6C	Home			(7)
			word	($ff << 8) | $ff	'6D
			word	($ff << 8) | $ff	'6E
			word	($ff << 8) | $ff	'6F
			word	((2*8+7)<<8) | (6*8+2)	'70	Insert			(0)
			word	((3*8+7)<<8) | (5*8+1)	'71	Delete			(.)
			word	((3*8+6)<<8) | (6*8+3)	'72	Down			(2)
			word	($ff << 8) | (1*8+2)	'73				(5)
			word	((4*8+6)<<8) | (2*8+2)	'74	Right			(6)
			word	((1*8+6)<<8) | (4*8+2)	'75	Up			(8)
			word	($ff << 8) | (5*8+6)	'76			Esc
			word	($ff << 8) | $ff	'77			NumLock
			word	($ff << 8) | $ff	'78			F11
			word	($ff << 8) | (5*8+0)	'79				(+)
			word	((5*8+7)<<8) | (5*8+3)	'7A	PageDn			(3)
			word	($ff << 8) | (0*8+2)	'7B				(-)
			word	($ff << 8) | (6*8+0)	'7C	PrScr			(*)
			word	((4*8+7)<<8) | (5*8+2)	'7D	PageUp			(9)
			word	($ff << 8) | $ff	'7E			ScrLock
			word	($ff << 8) | $ff	'7F
			word	($ff << 8) | $ff	'80
			word	($ff << 8) | $ff	'81
			word	($ff << 8) | $ff	'82
			word	($ff << 8) | $ff	'83			F7
'
'
' Uninitialized data
'
dmask			res	1
cmask			res	1
stat			res	1
data			res	1
x			res	1
y			res	1
t			res	1

_states			res	8	'write-only
_present		res	1	'write-only
_dpin			res	1	'read-only at start
_cpin			res	1	'read-only at start
_locks			res	1	'read-only at start
_auto			res	1	'read-only at start

{{

+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  | +------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    | |files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
}}
